(* A public domain RFC1951 implementation *)
structure Deflate :> Deflate =
struct

	exception Unknown_block_type
	exception Invalid_huffman_code
	exception Misformed_codelens

	(* get huffman encoded val *)	
	fun get_huffman_code iostate tree =
		if DeflateHuffman.has_step tree then
			get_huffman_code
				(DeflateIO.drop_bits iostate 1)
				(DeflateHuffman.step tree (hd (DeflateIO.take_bits iostate 1)))
		else
			(case DeflateHuffman.code tree []
			of SOME c => c
			| NONE => raise Invalid_huffman_code, iostate)  

	(* 
	 definition of deflate code/bits decoder that applies to 
	 both the mixed linear/length and distance alphabets 
		#1: huffman tree for initial code
		#2: get number of extra bits from initial code
		#3: get base for given code, for when there are extra bits
	*)

	type deflatedecoder = DeflateHuffman.tree * (int -> int) * (int -> int)

	fun decoder_tree (t,_,_) = t
	
	(* decode code, value *)
	fun get_code_value (hutree, extrafunc, basefunc) ioinit =
		let
			fun decoder_value code bitsval = (basefunc code) + DeflateBits.to_int_lsb_first bitsval
	
			val (code, iocode) = get_huffman_code ioinit hutree
			val nextra = extrafunc code
			val extra = DeflateIO.take_bits iocode nextra
		in
			((code, decoder_value code extra), DeflateIO.drop_bits iocode nextra)
		end 
	
	(* construct dedoder from tree *)

	fun literalen_decoder hutree = (
		(* huffman tree for literal/len alphabet *)
		hutree,
		(* code -> extra bits *)
		fn code => if code <= 264 orelse code = 285 then 0
				else if code <= 268 then 1
				else if code <= 272 then 2
				else if code <= 276 then 3
				else if code <= 280 then 4
				else 5,
		(* code -> extra bits base *)
		fn code => if code <= 264 then code - 257 + 3
				else if code <= 268 then (code - 265) * 2 + 11
				else if code <= 272 then (code - 269) * 4 + 19
				else if code <= 276 then (code - 273) * 8 + 35
				else if code <= 280 then (code - 277) * 16 + 67
				else if code <= 284 then (code - 281) * 32 + 131
				else 258)

	fun distance_decoder hutree = (
		(* huffman tree for code 0-32 *)
		hutree,
		(* code -> extra bits *)
		fn code => if code <= 3 then 0
				else if code >= 28 then 13
				else (code div 2) - 1,
		(* code -> base for extra bits *)
		fn code => if code <= 3 then code + 1
				else if code mod 2 = 0 then DeflateBits.pow2 (code div 2) + 1
				else DeflateBits.pow2 (code div 2) + 1 + DeflateBits.pow2 (code div 2 - 1))

	val fixed_decoder = 
		let 
			fun interval_replicate (fst, lst) = DeflateList.replicate (lst - fst + 1)
		in
			(literalen_decoder (DeflateHuffmanBuild.from_lengths 
				(interval_replicate (0, 143) 8 	
				@ interval_replicate (144, 255) 9
				@ interval_replicate (256, 279) 7
				@ interval_replicate (280, 287) 8)),
		 	distance_decoder 
	 			(DeflateHuffmanBuild.from_lengths (DeflateList.replicate 32 5)))
		end

	(* RFC 3.2.4 Non-compressed blocks *)

	fun uncompressed iostate =
		let
			val (len, iolen) = DeflateIO.get_fixed 16 (DeflateIO.drop_to_byte iostate)
			val (nlen, ionlen) = DeflateIO.get_fixed 16 iolen 
			(* FIXME check nlen! *)
			val ioadded = DeflateIO.append_bytes ionlen (DeflateIO.take_bytes ionlen len)
		in
			DeflateIO.drop_bytes ioadded len
		end

	(* RFC 3.2.6 Compression with fixed Huffman codes *)

	fun compressed (litlen_decoder,dist_decoder) ioinit = 
		let 
			val end_of_block = 256		
			val ((literallen, length), iolitlen) = get_code_value litlen_decoder ioinit
			
			val iodone = 
				if literallen = end_of_block then
					iolitlen
				else if literallen < end_of_block then
					DeflateIO.append_byte iolitlen literallen
				else
					let
						val ((distcode, distance), iodist) = get_code_value dist_decoder iolitlen
					in
						DeflateIO.append_from_output iodist length distance
					end
		in
			if literallen = end_of_block then iodone
			else compressed (litlen_decoder, dist_decoder) iodone
		end

	(* 3.2.7. Compression with dynamic Huffman codes *)

	fun huffman_reorder mixed =
		let
			val padded = mixed @ DeflateList.replicate 19 0
			fun nth x = List.nth (padded, x)
		in
			nth 3::nth 17::nth 15::nth 13::nth 11
				::nth 9::nth 7::nth 5::nth 4::nth 6
				::nth 8::nth 10::nth 12::nth 14
				::nth 16::nth 18::nth 0::nth 1::nth 2::nil
		end
		
	fun huffman_bitlens_decoder bitlens =
		(
		DeflateHuffmanBuild.from_lengths (huffman_reorder bitlens),
		(* extra bits *) 
		fn 16 => 2
		 | 17 => 3
		 | 18 => 7
		 | _ => 0,
		(* base for extra bits to int *)
		fn 16 => 3
		 | 17 => 3
		 | 18 => 11
		 | _ => 0
		)

	fun huffman_bitlens_expand codes =
		let
			fun f ((code,value), rs) =
				if code <= 15 then code :: rs
				else if code = 16 then (DeflateList.replicate value (hd rs)) @ rs
				else (DeflateList.replicate value 0) @ rs
		in
			rev (foldl f [] codes)
		end

	fun compressed_fixed s = compressed fixed_decoder s
	
	fun compressed_variable state1 =
		let
			fun get_codes decoder remcodes io =
				if remcodes < 0 then
					raise Misformed_codelens 
				else if remcodes = 0 then
					([], io)
				else
					let
						fun number_of_items (18,v) = v
						  | number_of_items (17,v) = v
						  | number_of_items (16,v) = v
						  | number_of_items _ = 1  
						val (code_value, ionext) = get_code_value decoder io
						val (result, ioresult) = get_codes decoder (remcodes - number_of_items code_value) ionext
					in
						(code_value::result, ioresult)
					end
			
			(* get the fixed sized lengths of the huffman dictionaries *)
			val (lens, state2) = 
				DeflateList.state_map 
					[DeflateIO.get_fixed 5, DeflateIO.get_fixed 5, DeflateIO.get_fixed 4] 
					state1
			val hlit_codes = List.nth (lens, 0) + 257
			val hdist_codes = List.nth (lens, 1) + 1
			val hclen_codes = List.nth (lens, 2) + 4

			(* get the bitlens defining the huffman dict for the real dictionaries *)
			val (hclen_bitlens, state3) = 
					DeflateList.state_map 
						(DeflateList.replicate hclen_codes (DeflateIO.get_fixed 3)) state2
			val huff_decoder = huffman_bitlens_decoder hclen_bitlens

			(* get the bitlens for both dictionaries using the dictionary above *)
			val (codes, state4) = get_codes huff_decoder (hlit_codes + hdist_codes) state3

			(* uncompress the bitlens and use them to build the 2 huffman dictionaries *)
			val hlit_hdist_bitlens = huffman_bitlens_expand codes
			val hlit_decoder = 
				literalen_decoder (DeflateHuffmanBuild.from_lengths (List.take (hlit_hdist_bitlens, hlit_codes)))
			val hdist_decoder = 
				distance_decoder (DeflateHuffmanBuild.from_lengths (List.drop (hlit_hdist_bitlens, hlit_codes)))
		in
			compressed (hlit_decoder, hdist_decoder) state4
		end

	(* RFC 3.2.3 Top level block format *)
	
	fun inflate ioinit =
  		let
			fun uncompress 0 io = uncompressed io
			  | uncompress 1 io = compressed fixed_decoder io
			  | uncompress 2 io = compressed_variable io
			  | uncompress _ _ = raise Unknown_block_type

			val last = hd (DeflateIO.take_bits ioinit 1)
			val (algo_type, ioalgo) = DeflateIO.get_fixed 2 (DeflateIO.drop_bits ioinit 1)
			val ionext = uncompress algo_type ioalgo
		in
			if last then ionext else inflate ionext
		end

	fun inflate_list [] = []
	  | inflate_list bs = DeflateIO.output (inflate (DeflateIO.from_input bs))

	(* self test *)
	val tests =
		let
			val ((litlen_tree, _, _), (dist_tree, _, _)) = fixed_decoder
		in [
		(* dependents tests *)
		List.all (fn x=>x) DeflateList.tests,
		List.all (fn x=>x) DeflateBits.tests,
		List.all (fn x=>x) DeflateIO.tests, 
		List.all (fn x=>x) DeflateHuffman.tests, 
		List.all (fn x=>x) DeflateHuffmanBuild.tests,
		
		(* check reordering function reorders as in 3.2.7 *)
		huffman_reorder [16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15] = DeflateList.enum 0 18,
		(* repeat coding of codelengths *)
		huffman_bitlens_expand [(17,3),(1,0),(15,0),(8,0),(16,3),(10,0)] = [0,0,0,1,15,8,8,8,8,10],
		
		(* extra bits *) 
		DeflateHuffman.collapse_string (decoder_tree (huffman_bitlens_decoder [2,2,0,1])) =
			[("0",0),("10",16),("11",17)],
		(* test that the construction of the fixed trees results in what 3.2.6 expects *)
		DeflateHuffman.collapse_summarised litlen_tree = [
			("00110000",0),("10111111",143),
			("110010000",144),("111111111",255),
			("0000000",256), ("0010111", 279), 
			("11000000",280), ("11000111",287)],
		DeflateHuffman.collapse_summarised dist_tree = [
			("00000",0),("11111",31)]
		]
		end

end
